import csv
from copy import deepcopy
from datetime import datetime
import jinja2
import json
import logging
import os
import re
import sys
from tabulate import tabulate


class Report:
    def __init__(self, queue, output):
        self.results = self._convert_q2list(queue)
        self.output = output
        self.logger = logging.getLogger('changeme')

    def render_csv(self,):
        fname = self.output
        if not re.match(r'.*\.csv$', fname):
            fname += ".csv"

        with open(fname, 'w') as fout:
            fieldnames = ["name", "username", "password", "target"]
            writer = csv.DictWriter(
                fout,
                quoting=csv.QUOTE_ALL,
                fieldnames=fieldnames,
                extrasaction='ignore'
            )
            writer.writeheader()
            writer.writerows(self.results)

        self.logger.critical("%i credentials written to %s" % (len(self.results), fname))

    def render_json(self):
        # convert the Target classes to a string so it can be json'd
        res = deepcopy(self.results)
        for r in res:
            t = r['target']
            r['target'] = str(t)

        results = dict()
        results["results"] = res
        j = json.dumps(results)
        fname = self.output
        if not re.match(r'.*\.json$', fname):
            fname += ".json"

        with open(fname, 'w') as fout:
            fout.write(j)

        self.logger.critical("%i credentials written to %s" % (len(self.results), fname))

    def print_results(self):
        if len(self.results) > 0:
            results = deepcopy(self.results)
            for r in results:
                if 'http' in r['target'].protocol:
                    r['evidence'] = ''

            print("")
            print("")
            self.logger.critical('Found %i default credentials' % len(self.results))
            print("")
            print(tabulate(results, headers={'name': 'Name',
                                             'username': 'Username',
                                             'password': 'Password',
                                             'target': 'Target',
                                             'evidence': 'Evidence'}))

            print("")
        else:
            print("No default credentials found")

    def render_html(self):
        template_loader = jinja2.FileSystemLoader(searchpath=self.get_template_path())
        template_env = jinja2.Environment(loader=template_loader)
        report_template = template_env.get_template('report.j2')
        cli = ' '.join(sys.argv)
        timestamp = datetime.now()
        report = report_template.render({'found': self.results, 'cli': cli, 'timestamp': timestamp})

        fname = self.output
        if not re.match(r'.*\.html$', fname):
            fname += ".html"

        with open(fname, 'w') as fout:
            fout.write(report)

        self.logger.critical("%i credentials written to %s" % (len(self.results), fname))

    @staticmethod
    def get_template_path():
        PATH = os.path.dirname(os.path.abspath(__file__))
        template_path = os.path.join(PATH, 'templates')
        return template_path

    def _convert_q2list(self, q):
        items = list()
        while not q.qsize() == 0:
            i = q.get()
            items.append(i)

        # Restore queue
        for i in items:
            q.put(i)

        return items



from .database import Database

class MSSQL(Database):
    def __init__(self, cred, target, username, password, config):
        super(MSSQL, self).__init__(cred, target, username, password, config)
        self.target.protocol = "mssql+pyodbc"
        self.database = ""
        self.query = "SELECT @@VERSION AS 'SQL Server Version';"

    def _mkscanner(self, cred, target, u, p, config):
        return MSSQL(cred, target, u, p, config)


__version__ = '1.2.1'
contributors = [
    "ztgrace",
    "the-c0d3r",
    "Graph-X",
    "AlessandroZ",
    "ThomasTJ",
    "Alistair Chapman",
    "John Van de Meulebrouck Brendgard",
    "network23",
    "decidedlygray",
    "Joe Testa",
    "Chandrapal",
    "Naglis Jonaitis",
    "Samuel Henrique",
    "sil3ntcor3",
]



import base64
import random
from requests import session
from .scanner import Scanner
import re
from selenium import webdriver
from time import sleep
try:
    # Python 3
    from urllib.parse import urlencode, urlparse
except ImportError:
    # Python 2
    from urllib import urlencode
    from urlparse import urlparse

HEADERS_USERAGENTS = [
    'Mozilla/5.0 (X11; U; Linux x86_64; en-US; rv:1.9.1.3) Gecko/20090913 Firefox/3.5.3',
    'Mozilla/5.0 (Windows; U; Windows NT 6.1; en; rv:1.9.1.3) Gecko/20090824 Firefox/3.5.3 (.NET CLR 3.5.30729)',
    'Mozilla/5.0 (Windows; U; Windows NT 5.2; en-US; rv:1.9.1.3) Gecko/20090824 Firefox/3.5.3 (.NET CLR 3.5.30729)',
    'Mozilla/5.0 (Windows; U; Windows NT 6.1; en-US; rv:1.9.1.1) Gecko/20090718 Firefox/3.5.1',
    'Mozilla/5.0 (Windows; U; Windows NT 5.1; en-US) AppleWebKit/532.1 (KHTML, like Gecko) Chrome/4.0.219.6 Safari/532.1',
    'Mozilla/4.0 (compatible; MSIE 8.0; Windows NT 6.1; WOW64; Trident/4.0; SLCC2; .NET CLR 2.0.50727; InfoPath.2)',
    'Mozilla/4.0 (compatible; MSIE 8.0; Windows NT 5.2; Win64; x64; Trident/4.0)',
    'Mozilla/4.0 (compatible; MSIE 8.0; Windows NT 5.1; Trident/4.0; SV1; .NET CLR 2.0.50727; InfoPath.2)',
    'Mozilla/5.0 (Windows; U; MSIE 7.0; Windows NT 6.0; en-US)',
    'Mozilla/4.0 (compatible; MSIE 6.1; Windows XP)',
    'Opera/9.80 (Windows NT 5.2; U; ru) Presto/2.5.22 Version/10.51'
]


def get_useragent():
    return random.choice(HEADERS_USERAGENTS)


class HTTPGetScanner(Scanner):

    def __init__(self, cred, target, username, password, config, cookies):
        super(HTTPGetScanner, self).__init__(cred, target, config, username, password)
        self.cred = cred
        self.config = config
        self.cookies = cookies
        self.headers = dict()
        self.request = session()
        self.response = None

        headers = self.cred['auth'].get('headers', dict())
        custom_ua = False
        if headers:
            for h in headers:
                self.headers.update(h)
                if not custom_ua and any(k.lower() == 'user-agent' for k in h):
                    custom_ua = True

        # If set, take user agent from CLI args, otherwise, pick a random
        # one if not provided in the cred file.
        if self.config.useragent:
            self.headers.update(self.config.useragent)
        elif not custom_ua:
            self.headers.update({'User-Agent': get_useragent()})

        # make the cred have only one u:p combo
        self.cred['auth']['credentials'] = [{'username': self.username, 'password': self.password}]

    def __reduce__(self):
        return self.__class__, (self.cred, self.target, self.username, self.password, self.config, self.cookies)

    def scan(self):
        try:
            self._make_request()
        except Exception as e:
            self.logger.error('Failed to connect to %s' % self.target)
            self.logger.debug('Exception: %s: %s' % (type(e).__name__, e.__str__().replace('\n', '|')))
            return None

        if self.response.status_code == 429:
            self.warn('Status 429 received. Sleeping for %d seconds and trying again' % self.config.delay)
            sleep(self.config.delay)
            try:
                self._make_request()
            except Exception as e:
                self.logger.error('Failed to connect to %s' % self.target)

        return self.check_success()

    def check_success(self):
        match = False
        success = self.cred['auth']['success']

        if self.cred['auth'].get('base64', None):
            self.username = base64.b64decode(self.cred.username)
            self.password = base64.b64decode(self.cred.password)

        if success.get('status') == self.response.status_code or self.response.history and self.response.history[0].status_code == success.get('status'):
            self.logger.debug('%s matched %s success status code %s' % (self.target, self.cred['name'], self.response.status_code))
            if success.get('body'):
                for string in success.get('body'):
                    if re.search(string, self.response.text, re.IGNORECASE):
                        self.logger.debug('%s matched %s success body text %s' % (self.target, self.cred['name'], success.get('body')))
                        match = True
                        break
            else:
                match = True

        if match:
            self.logger.critical('[+] Found %s default cred %s:%s at %s' %
                                 (self.cred['name'], self.username, self.password, self.target))
            evidence = ''
            if self.config.output is not None:
                try:
                    evidence = self._screenshot(self.target)
                except Exception as e:
                    self.logger.error("Error gathering screenshot for %s" % self.target)
                    self.logger.debug('Exception: %s: %s' % (type(e).__name__, e.__str__().replace('\n', '|')))

            return {'name': self.cred['name'],
                    'username': self.username,
                    'password': self.password,
                    'target': self.target,
                    'evidence': evidence}
        else:
            self.logger.info('Invalid %s default cred %s:%s at %s' %
                             (self.cred['name'], self.username, self.password, self.target))
            return False

    def _check_fingerprint(self):
        self.logger.debug("_check_fingerprint")
        self.request = session()
        self.response = self.request.get(self.target,
                                         timeout=self.config.timeout,
                                         verify=False,
                                         proxies=self.config.proxy,
                                         cookies=self.fingerprint.cookies,
                                         headers=self.fingerprint.headers)
        self.logger.debug('_check_fingerprint', '%s - %i' % (self.target, self.response.status_code))
        return self.fingerprint.match(self.response)

    def _make_request(self):
        self.logger.debug("_make_request")
        data = self.render_creds(self.cred)
        qs = urlencode(data)
        url = "%s?%s" % (self.target, qs)
        self.logger.debug("url: %s" % url)
        self.response = self.request.get(self.target,
                                         verify=False,
                                         proxies=self.config.proxy,
                                         timeout=self.config.timeout,
                                         headers=self.headers,
                                         cookies=self.cookies)

    def render_creds(self, candidate, csrf=None):
        """
            Return a list of dicts with post/get data and creds.

            The list of dicts have a data element and a username and password
            associated with the data. The data will either be a dict if its a
            regular GET or POST and a string if its a raw POST.
        """
        b64 = candidate['auth'].get('base64', None)
        type = candidate['auth'].get('type')
        config = None
        if type == 'post':
            config = candidate['auth'].get('post', None)
        if type == 'get':
            config = candidate['auth'].get('get', None)

        if not type == 'raw_post':
            data = self._get_parameter_dict(candidate['auth'])

            if csrf:
                csrf_field = candidate['auth']['csrf']
                data[csrf_field] = csrf

            for cred in candidate['auth']['credentials']:
                cred_data = {}
                username = ""
                password = ""
                if b64:
                    username = base64.b64encode(cred['username'])
                    password = base64.b64encode(cred['password'])
                else:
                    username = cred['username']
                    password = cred['password']

                cred_data[config['username']] = username
                cred_data[config['password']] = password

                data_to_send = dict(list(data.items()) + list(cred_data.items()))
                return data_to_send
        else:  # raw post
            return None

    def _get_parameter_dict(self, auth):
        params = dict()
        data = auth.get('post', auth.get('get', None))
        for k in list(data.keys()):
            if k not in ('username', 'password', 'url'):
                params[k] = data[k]

        return params

    @staticmethod
    def get_base_url(req):
        parsed = urlparse(req)
        url = "%s://%s" % (parsed[0], parsed[1])
        return url

    def _screenshot(self, target):
        self.logger.debug("Screenshotting %s" % self.target)
        # Set up the selenium webdriver
        # This feels like it will have threading issues
        for key, value in self.response.request.headers.items():
            capability_key = 'phantomjs.page.customHeaders.{}'.format(key)
            webdriver.DesiredCapabilities.PHANTOMJS[capability_key] = value

        if self.config.proxy:
            webdriver.DesiredCapabilities.PHANTOMJS['proxy'] = {
                    "httpProxy": self.config.proxy['http'].replace('http://', ''),
                    "ftpProxy": self.config.proxy['http'].replace('http://', ''),
                    "sslProxy": self.config.proxy['http'].replace('http://', ''),
                    "noProxy":None,
                    "proxyType":"MANUAL",
                    "autodetect":False
            }
        driver = webdriver.PhantomJS()
        driver.set_page_load_timeout(int(self.config.timeout) - 0.1)
        driver.set_window_position(0, 0)
        driver.set_window_size(850, 637.5)
        for cookie in self.response.request._cookies.items():
            self.logger.debug("Adding cookie: %s:%s" % cookie)
            driver.add_cookie({'name': cookie[0],
                               'value': cookie[1],
                               'path': '/',
                               'domain': self.target.host
            })

        try:
            driver.get(str(self.target))
            driver.save_screenshot('screenshot.png')
            evidence = driver.get_screenshot_as_base64()
            driver.quit()
        except Exception as e:
            self.logger.error('Error getting screenshot for %s' % self.target)
            self.logger.debug('Exception: %s: %s' % (type(e).__name__, e.__str__().replace('\n', '|')))
            evidence = ""

        return evidence




import logging
import multiprocessing as mp
import redis
from changeme.redis_queue import RedisQueue
import pickle
from .scanners.ftp import FTP
from .scanners.http_fingerprint import HttpFingerprint
from .scanners.memcached import MemcachedScanner
from .scanners.mongo import Mongodb
from .scanners.mssql import MSSQL
from .scanners.mysql import MySQL
from .scanners.postgres import Postgres
from .scanners.redis_scanner import RedisScanner
from .scanners.snmp import SNMP
from .scanners.ssh import SSH
from .scanners.ssh_key import SSHKey
from .scanners.telnet import Telnet
from .scanners.http_fingerprint import HttpFingerprint
from .target import Target
import time
try:
    # Python 2
    from Queue import Queue
except:
    # Python 3
    from queue import Queue


class ScanEngine(object):
    def __init__(self, creds, config):
        self.creds = creds
        self.config = config
        self.logger = logging.getLogger('changeme')
        self.scanners = self._get_queue('scanners')
        self.total_scanners = 0
        self.targets = set()
        self.fingerprints = self._get_queue('fingerprints')
        self.total_fps = 0
        self.found_q = self._get_queue('found_q')

    def scan(self):

        # Phase I - Fingerprint
        ######################################################################
        if not self.config.resume:
            self._build_targets()

        if self.config.dryrun:
            self.dry_run()

        num_procs = self.config.threads if self.fingerprints.qsize() > self.config.threads else self.fingerprints.qsize()

        self.logger.debug('Number of procs: %i' % num_procs)
        self.total_fps = self.fingerprints.qsize()
        procs = [mp.Process(target=self.fingerprint_targets) for i in range(num_procs)]

        self._add_terminators(self.fingerprints)

        for proc in procs:
            proc.start()

        for proc in procs:
            proc.join()

        self.logger.info('Fingerprinting completed')

        # Phase II - Scan
        ######################################################################
        # Unique the queue
        scanners = list()
        while self.scanners.qsize() > 0:
            s = self.scanners.get()

            if s not in scanners:
                scanners.append(s)

        for s in scanners:
            self.scanners.put(s)

        if not self.config.fingerprint:
            num_procs = self.config.threads if self.scanners.qsize() > self.config.threads else self.scanners.qsize()
            self.total_scanners = self.scanners.qsize()

            self.logger.debug('Starting %i scanner procs' % num_procs)
            procs = [mp.Process(target=self._scan, args=(self.scanners, self.found_q)) for i in range(num_procs)]

            self._add_terminators(self.scanners)

            for proc in procs:
                self.logger.debug('Starting scanner proc')
                proc.start()

            for proc in procs:
                proc.join()

            self.logger.info('Scanning Completed')

            # Hack to address a broken pipe IOError per https://stackoverflow.com/questions/36359528/broken-pipe-error-with-multiprocessing-queue
            time.sleep(0.1)

    def _add_terminators(self, q):
        # Add poison pills
        for i in range(self.config.threads):
            q.put(None)

    def _scan(self, scanq, foundq):
        while True:
            remaining = self.scanners.qsize()
            self.logger.debug('%i scanners remaining' % remaining)

            try:
                scanner = scanq.get(block=True)
                if scanner is None:
                    return
            except Exception as e:
                self.logger.debug('Caught exception: %s' % type(e).__name__)
                continue

            result = scanner.scan()
            if result:
                foundq.put(result)

    def fingerprint_targets(self):
        while True:
            remaining = self.fingerprints.qsize()
            self.logger.debug('%i fingerprints remaining' % remaining)

            try:
                fp = self.fingerprints.get()
                if type(fp) == bytes:
                    fp = pickle.loads(fp)

                # Exit process
                if fp is None:
                    return

            except Exception as e:
                self.logger.debug('Caught exception: %s' % type(e).__name__)
                self.logger.debug('Exception: %s: %s' % (type(e).__name__, e.__str__().replace('\n', '|')))
                return

            if fp.fingerprint():
                results = fp.get_scanners(self.creds)
                if results:
                    for result in results:
                        self.scanners.put(result)
            else:
                self.logger.debug('failed fingerprint')

        self.logger.debug('scanners: %i, %s' % (self.scanners.qsize(), id(self.scanners)))

    def _build_targets(self):
        self.logger.debug('Building targets')

        if self.config.target:
            self.targets = Target.parse_target(self.config.target)
        else:
            self.logger.warning('shodan')
            self.targets = Target.get_shodan_targets(self.config)


        # Load set of targets into queue
        self.logger.debug('%i targets' % len(self.targets))

        # If there's only one protocol and the user specified a protocol, override the defaults
        if len(self.targets) == 1:
            t = self.targets.pop()
            if t.protocol:
                self.config.protocols = t.protocol
            self.targets.add(t)

        fingerprints = list()
        # Build a set of unique fingerprints
        if 'http' in self.config.protocols or self.config.all:
            fingerprints = fingerprints + HttpFingerprint.build_fingerprints(self.targets, self.creds, self.config)

        fingerprints = list(set(fingerprints))  # unique the HTTP fingerprints

        # Add any protocols if they were included in the targets
        for t in self.targets:
            if t.protocol and t.protocol not in self.config.protocols:
                self.config.protocols += ",%s" % t.protocol

        self.logger.info('Configured protocols: %s' % self.config.protocols)

        # scanner_map maps the friendly proto:// name to the actual class name
        scanner_map = {
            'ssh': 'SSH',
            'ssh_key': 'SSHKey',
            'ftp': 'FTP',
            'memcached': 'MemcachedScanner',
            'mongodb': 'Mongodb',
            'mssql': 'MSSQL',
            'mysql': 'MySQL',
            'postgres': 'Postgres',
            'redis': 'RedisScanner',
            'snmp': 'SNMP',
            'telnet': 'Telnet',
        }

        for target in self.targets:
            for cred in self.creds:
                for proto, classname in scanner_map.items():
                    if cred['protocol'] == proto and (proto in self.config.protocols or self.config.all):
                        t = Target(host=target.host, port=target.port, protocol=proto)
                        fingerprints.append(globals()[classname](cred, t, self.config, '', ''))

        self.logger.info("Loading creds into queue")
        for fp in set(fingerprints):
            self.fingerprints.put(fp)
        self.total_fps = self.fingerprints.qsize()
        self.logger.debug('%i fingerprints' % self.fingerprints.qsize())


    def dry_run(self):
        self.logger.info("Dry run targets:")
        while self.fingerprints.qsize() > 0:
            fp = self.fingerprints.get()
            print(fp.target)
        quit()

    def _get_queue(self, name):
        try:
            # Try for redis
            r = RedisQueue(name)
            r.ping()
            self.logger.debug('Using RedisQueue for %s' % name)
            return r

        except redis.ConnectionError:
            # Fall back to sqlite persistent queue
            self.logger.debug('Using in-memory queue for %s' % name)
            m = mp.Manager()
            q = m.Queue()
            return q



import redis
from .scanner import Scanner


class RedisScanner(Scanner):

    def __init__(self, cred, target, username, password, config):
        super(RedisScanner, self).__init__(cred, target, config, username, password)

    def _check(self):
        r = redis.StrictRedis(host=self.target.host, port=self.target.port)
        info = r.info()
        evidence = "redis_version: %s, os: %s" % (info['redis_version'], info['os'])

        return evidence

    def _mkscanner(self, cred, target, u, p, config):
        return RedisScanner(cred, target, u, p, config)



from pysnmp.hlapi import *
from .scanner import Scanner


class SNMP(Scanner):
    def __init__(self, cred, target, username, password, config):
        super(SNMP, self).__init__(cred, target, config, username, password)

    def fingerprint(self):
        # Don't fingerprint since it's UDP
        return True

    def _check(self):
        iterator = getCmd(SnmpEngine(),
                          CommunityData(self.password),
                          UdpTransportTarget((str(self.target.host), 161)),
                          ContextData(),
                          ObjectType(ObjectIdentity('SNMPv2-MIB', 'sysDescr', 0)))

        errorIndication, errorStatus, errorIndex, varBinds = next(iterator)

        evidence = ""
        if errorIndication:
            self.logger.debug(errorIndication)
        elif errorStatus:
            self.logger.debug('%s at %s' % (errorStatus.prettyPrint(),
                                errorIndex and varBinds[int(errorIndex) - 1][0] or '?'))
        else:
            for varBind in varBinds:
                evidence += ' = '.join([x.prettyPrint() for x in varBind])

        if evidence == "":
            raise Exception

        return evidence

    def _mkscanner(self, cred, target, u, p, config):
        return SNMP(cred, target, u, p, config)



import argparse
from changeme.scan_engine import ScanEngine
from changeme.target import Target
from changeme import core
from .core import cli_args
from copy import deepcopy
import csv
import json
import logging
import mock
from .mock_responses import MockResponses
from nose.tools import *
import os
import responses

"""
TODO:
 - Custom headers
 - 429 response code
 -

"""

def reset_handlers():
    logger = logging.getLogger('changeme')
    logger.handlers = []
    core.remove_queues()
    

fp_args = deepcopy(cli_args)
fp_args['nmap'] = 'tests/tomcat_nmap.xml'
fp_args['name'] = 'Tomcat'
@responses.activate
@mock.patch('argparse.ArgumentParser.parse_args', return_value=argparse.Namespace(**fp_args))
def test_tomcat_match_nmap(mock_args):
    def tomcat_callback(request):
        if request.headers.get('Authorization', False):
            return (200, MockResponses.tomcat_auth['adding_headers'], MockResponses.tomcat_auth['body'])
        else:
            return (401, MockResponses.tomcat_fp['adding_headers'], '')

    responses.add_callback(
        responses.GET,
        MockResponses.tomcat_fp['url'],
        callback=tomcat_callback,
    )

    reset_handlers()
    try:
        os.remove(core.PERSISTENT_QUEUE)
    except OSError:
        pass

    args = core.parse_args()
    core.init_logging(args['args'].verbose, args['args'].debug, args['args'].log)
    config = core.Config(args['args'], args['parser'])
    creds = core.load_creds(config)
    s = ScanEngine(creds, config)
    s._build_targets()
    s._add_terminators(s.fingerprints)

    print(("fp: %i" % s.fingerprints.qsize()))
    s.fingerprint_targets()

    # Queue is not serializeable so we can't copy it using deepcopy
    scanners = list()
    print(("scanners: %s" % s.scanners.qsize()))

    t1 = Target(host='127.0.0.1', port=8080, protocol='http', url='/manager/html')
    t2 = Target(host='127.0.0.1', port=8080, protocol='http', url='/tomcat/manager/html')
    while s.scanners.qsize() > 0:
        scanner = s.scanners.get()
        assert scanner.target == t1 or scanner.target == t2
        scanners.append(scanner)

    # Load the scanners back into the queue
    for scanner in scanners:
        s.scanners.put(scanner)
    assert s.scanners.qsize() == 34
    s._add_terminators(s.scanners)

    responses.reset()
    responses.add(**MockResponses.tomcat_auth)
    s._scan(s.scanners, s.found_q)
    assert s.found_q.qsize() == 17


fp_args = deepcopy(cli_args)
fp_args['fingerprint'] = True
fp_args['name'] = 'Tomcat'
fp_args['noversion'] = False
@responses.activate
@mock.patch('argparse.ArgumentParser.parse_args', return_value=argparse.Namespace(**fp_args))
def test_tomcat_fingerprint(mock_args):
    responses.add(**MockResponses.tomcat_fp)
    reset_handlers()
    se = core.main()
    print(("Scanners:",se.scanners.qsize()))
    assert se.scanners.qsize() == 34
    core.remove_queues()

@responses.activate
@mock.patch('argparse.ArgumentParser.parse_args', return_value=argparse.Namespace(**cli_args))
def test_tomcat_invalid_creds(mock_args):
    responses.add(**MockResponses.tomcat_fp)
    reset_handlers()
    se = core.main()
    assert se.found_q.qsize() == 0

@responses.activate
@mock.patch('argparse.ArgumentParser.parse_args', return_value=argparse.Namespace(**cli_args))
def test_jboss_scan_fail(mock_args):
    responses.add(**MockResponses.jboss_fp)
    responses.add(**MockResponses.jboss_auth_fail)

    reset_handlers()
    args = core.parse_args()
    core.init_logging(args['args'].verbose, args['args'].debug, args['args'].log)
    config = core.Config(args['args'], args['parser'])
    creds = core.load_creds(config)
    se = ScanEngine(creds, config)
    se._build_targets()
    se._add_terminators(se.fingerprints)
    se.fingerprint_targets()
    print(se.scanners.qsize())
    scanners = list()
    while se.scanners.qsize() > 0:
        s = se.scanners.get()
        print(s.cred['name'])
        print(s.target)
        print(s.username)
        print(s.password)
        scanners.append(s)

    print("num scanners: %i" % len(scanners))
    assert len(scanners) == 2

    # put scanners back in queue
    for s in scanners:
        se.scanners.put(s)

    se._add_terminators(se.scanners)
    se._scan(se.scanners, se.found_q)
    assert se.found_q.qsize() == 0


jboss_args = deepcopy(cli_args)
jboss_args['name'] = 'JBoss AS 6 Alt'
@responses.activate
@mock.patch('argparse.ArgumentParser.parse_args', return_value=argparse.Namespace(**jboss_args))
def test_jboss_scan_success(mock_args):
    responses.add(**MockResponses.jboss_fp)
    responses.add(**MockResponses.jboss_auth)
    reset_handlers()
    se = core.main()
    assert se.found_q.qsize() == 1


subnet_args = deepcopy(cli_args)
subnet_args['target'] = '127.0.0.1/32'
subnet_args['protocols'] = 'http'
subnet_args['name'] = 'JBoss AS 6 Alt'
@responses.activate
@mock.patch('argparse.ArgumentParser.parse_args', return_value=argparse.Namespace(**subnet_args))
def test_jboss_scan_success_subnet(mock_args):
    responses.add(**MockResponses.jboss_fp)
    responses.add(**MockResponses.jboss_auth)
    reset_handlers()
    se = core.main()
    assert se.found_q.qsize() == 1


@responses.activate
@mock.patch('argparse.ArgumentParser.parse_args', return_value=argparse.Namespace(**cli_args))
def test_jboss_csrf_fail(mock_args):
    responses.add(**MockResponses.jboss_fp_no_csrf)
    reset_handlers()
    se = core.main()
    assert se.found_q.qsize() == 0


idrac_args = deepcopy(cli_args)
idrac_args['name'] = "Dell iDRAC"
@responses.activate
@mock.patch('argparse.ArgumentParser.parse_args', return_value=argparse.Namespace(**idrac_args))
def test_idrac_scan_success(mock_args):
    responses.reset()
    responses.add(**MockResponses.idrac_fp)
    responses.add(**MockResponses.idrac_auth)
    reset_handlers()
    se = core.main()
    assert se.found_q.qsize() == 1


targets_args = deepcopy(cli_args)
targets_args['target'] = '/tmp/targets.txt'
@responses.activate
@mock.patch('argparse.ArgumentParser.parse_args', return_value=argparse.Namespace(**targets_args))
def test_targets_scan_success(mock_args):
    responses.reset()
    responses.add(**MockResponses.idrac_fp)
    responses.add(**MockResponses.idrac_auth)
    with open(targets_args['target'], 'w') as fout:
        fout.write('127.0.0.1' + '\n')

    reset_handlers()
    se = core.main()
    assert se.found_q.qsize() == 1


csv_args = deepcopy(cli_args)
csv_args['log'] = '/tmp/output.log'
csv_args['output'] = '/tmp/output.csv'
csv_args['name'] = 'JBoss AS 6 Alt'
@responses.activate
@mock.patch('argparse.ArgumentParser.parse_args', return_value=argparse.Namespace(**csv_args))
def test_csv_output(mock_args):
    responses.add(**MockResponses.jboss_fp)
    responses.add(**MockResponses.jboss_auth)
    reset_handlers()
    se = core.main()
    print(se.found_q.qsize())
    assert se.found_q.qsize() == 1

    assert os.path.isfile(csv_args['output'])
    i = 0
    with open(csv_args['output'], 'r') as csvfile:
        reader = csv.reader(csvfile)
        for line in reader:
            if i == 1:
                assert line[0] == 'JBoss AS 6 Alt'
                assert line[1] == 'admin'
                assert line[2] == 'admin'
                assert line[3] == 'http://127.0.0.1:8080/admin-console/login.seam'
            i += 1

    assert os.path.isfile(csv_args['log'])


json_args = deepcopy(cli_args)
json_args['output'] = '/tmp/output.json'
json_args['name'] = 'JBoss AS 6 Alt'
@responses.activate
@mock.patch('argparse.ArgumentParser.parse_args', return_value=argparse.Namespace(**json_args))
def test_json_output(mock_args):
    responses.add(**MockResponses.jboss_fp)
    responses.add(**MockResponses.jboss_auth)
    reset_handlers()
    se = core.main()
    assert se.found_q.qsize() == 1

    assert os.path.isfile(json_args['output'])
    i = 0
    with open(json_args['output'], 'r') as json_file:
        j = json.loads(json_file.read())
        assert j["results"][0]['name']      == 'JBoss AS 6 Alt'
        assert j['results'][0]['username']  == 'admin'
        assert j['results'][0]['password']  == 'admin'
        assert j['results'][0]['target']    == 'http://127.0.0.1:8080/admin-console/login.seam'


dr_args = deepcopy(cli_args)
dr_args['dryrun'] = True
@raises(SystemExit)
@mock.patch('argparse.ArgumentParser.parse_args', return_value=argparse.Namespace(**dr_args))
def test_dryrun(mock_args):
    reset_handlers()
    se = core.main()
    assert se.found_q.qsize() == 0


es_args = deepcopy(cli_args)
es_args['name'] = "elasticsearch"
@responses.activate
@mock.patch('argparse.ArgumentParser.parse_args', return_value=argparse.Namespace(**es_args))
def test_es_scan_success(mock_args):
    responses.reset()
    responses.add(**MockResponses.elasticsearch)
    reset_handlers()
    se = core.main()
    assert se.found_q.qsize() == 1



from changeme.scanners.http_basic_auth import HTTPBasicAuthScanner
from changeme.scanners.http_get import HTTPGetScanner
from changeme.scanners.http_post import HTTPPostScanner
from changeme.scanners.http_raw_post import HTTPRawPostScanner
from changeme.target import Target
from copy import deepcopy
import logging
from lxml import html
import re
import requests


class HttpFingerprint:
    def __init__(self, target, headers, cookies, config):
        self.target = target # changeme.target.Target()
        self.headers = headers
        self.cookies = cookies
        self.config = config
        self.logger = logging.getLogger('changeme')
        self.res = None
        self.req = requests.Session()

    def __getstate__(self):
        state = self.__dict__
        state['logger'] = None  # Need to clear the logger when serializing otherwise mp.Queue blows up
        return state

    def __setstate__(self, d):
        self.__dict__ = d
        self.logger = logging.getLogger('changeme')

    def __hash__(self):
        return hash(str(self.target) + str(self.headers) + str(self.cookies))

    def __eq__(self, other):
        s = dict()
        o = dict()
        s['target'] = self.target
        s['headers'] = self.headers
        s['cookies'] = self.cookies
        o['target'] = other.target
        o['headers'] = other.headers
        o['cookies'] = other.cookies
        return s == o

    def fingerprint(self):

        try:
            self._fp()
        except Exception as e:
            if self.config.ssl and e.__class__ == requests.exceptions.SSLError:
                self.target.protocol = 'http'
                self.logger.debug('Retrying with non-SSL target: %s' % self.target)
                try:
                    self._fp()
                except Exception as e:
                    self.logger.debug('Failed to connect to %s' % self.target)

            return False

        return True

    def _fp(self):
        self.res = self.req.get(
            str(self.target),
            timeout=self.config.timeout,
            verify=False,
            proxies=self.config.proxy,
            headers=self.headers,
            cookies=self.cookies
        )

    def _get_csrf_token(self, res, cred):
        name = cred['auth'].get('csrf', False)
        if name:
            tree = html.fromstring(res.content)
            try:
                csrf = str(tree.xpath('//input[@name="%s"]/@value' % name)[0])
            except:
                self.logger.error(
                    'Failed to get CSRF token %s in %s' % (str(name), str(res.url)))
                return False
            self.logger.debug('Got CSRF token %s: %s' % (name, csrf))
        else:
            csrf = False

        return csrf

    def _get_session_id(self, res, cred):
        cookie = cred['auth'].get('sessionid', False)

        if cookie:
            try:
                value = res.cookies[cookie]
                self.logger.debug('Got session cookie value: %s' % value)
            except:
                self.logger.error(
                    'Failed to get %s cookie from %s' % (cookie, res.url))
                return False
            return {cookie: value}
        else:
            self.logger.debug('No cookie')
            return False

    def ismatch(self, cred, response):
        match = False
        if cred['protocol'] == 'http':
            fp = cred['fingerprint']
            basic_auth = fp.get('basic_auth_realm', None)
            if basic_auth and basic_auth in response.headers.get('WWW-Authenticate', list()):
                self.logger.info('%s basic auth matched: %s' % (cred['name'], basic_auth))
                match = True

            server = response.headers.get('Server', None)
            fp_server = fp.get('server_header', None)
            if fp_server and server and fp_server in server:
                self.logger.debug('%s server header matched: %s' % (cred['name'], fp_server))
                match = True

            body = fp.get('body', None)
            if body:
                for b in body:
                    if re.search(b, response.text):
                        match = True
                        self.logger.info('%s body matched: %s' % (cred['name'], b))
                    elif body:
                        match = False

        return match

    def get_scanners(self, creds):
        scanners = list()
        for cred in creds:
            if self.ismatch(cred, self.res):

                csrf = self._get_csrf_token(self.res, cred)
                if cred['auth'].get('csrf', False) and not csrf:
                    self.logger.error('Missing required CSRF token')
                    return

                sessionid = self._get_session_id(self.res, cred)
                if cred['auth'].get('sessionid') and not sessionid:
                    self.logger.error("Missing session cookie %s for %s" % (cred['auth'].get('sessionid'), self.res.url))
                    return

                for pair in cred['auth']['credentials']:
                    for u in cred['auth']['url']:  # pass in the auth url
                        target = deepcopy(self.target)
                        target.url = u
                        self.logger.debug('Building %s %s:%s, %s' % (cred['name'], pair['username'], pair['password'], target))

                        if cred['auth']['type'] == 'get':
                            scanners.append(HTTPGetScanner(cred, target, pair['username'], pair['password'], self.config, self.req.cookies))
                        elif cred['auth']['type'] == 'post':
                            scanners.append(HTTPPostScanner(cred, target, pair['username'], pair['password'], self.config, self.req.cookies, csrf))
                        elif cred['auth']['type'] == 'raw_post':
                            scanners.append(HTTPRawPostScanner(cred, target, pair['username'], pair['password'], self.config, self.req.cookies, csrf, pair['raw']))
                        elif cred['auth']['type'] == 'basic_auth':
                            scanners.append(HTTPBasicAuthScanner(cred, target, pair['username'], pair['password'], self.config, self.req.cookies))

        return scanners

    @staticmethod
    def build_fingerprints(targets, creds, config):
        fingerprints = list()
        logger = logging.getLogger('changeme')
        # Build a set of unique fingerprints
        for target in targets:
            for c in creds:
                if not c['protocol'] == 'http':
                    continue
                if not config.portoverride and (target.port and not c['default_port'] == target.port):
                    continue

                fp = c['fingerprint']
                for url in fp.get('url'):
                    t = Target(host=target.host, port=target.port, protocol=target.protocol)
                    if c.get('ssl') or config.ssl:
                        t.protocol = 'https'
                    else:
                        t.protocol = 'http'

                    if not t.port:
                        t.port = c['default_port']
                    t.url = url

                    hfp = HttpFingerprint(t, fp.get('headers', None), fp.get('cookie', None), config)
                    logger.debug('Adding %s to fingerprint list' % hfp.target)
                    fingerprints.append(hfp)

        return fingerprints



#!/usr/bin/env python

import changeme.core
import os
try:
    # Python 3
    from urllib.parse import unquote_plus
except ImportError:
    # Python 2
    from urllib import unquote_plus
import yaml

cli_prompt = input
try:
    cli_prompt = raw_input
except NameError:
    pass

http_schema = {
    'auth': {
        'type': 'dict',
        'required': True,
        'schema': {
            'credentials': {
                'type': 'list',
                'required': True,
                'schema': {
                    'type': 'dict',
                    'schema': {
                        'username': {
                            'type': ['string', 'integer'],
                            'nullable': True,
                            'required': True,
                        },
                        'password': {
                            'type': ['string', 'integer'],
                            'nullable': True,
                            'required': True,
                        },
                        'ref': {'type': 'string', 'required': False},
                        'raw': {'type': 'string', 'required': False},
                    }
                }
            },
            'headers': {
                'type': 'list',
                'required': False,
                'schema': {
                    'type': 'dict'
                }
            },
            'csrf': {
                'type': 'string',
                'nullable': True,
                'required': False,
            },
            'post': {
                'type': 'dict',
                'allow_unknown': True,
                'schema': {
                    'username': {'type': 'string', 'required': True},
                    'password': {'type': 'string', 'required': True},
                }
            },
            'get': {
                'type': 'dict',
                'allow_unknown': True,
                'schema': {
                    'username': {'type': 'string', 'required': True},
                    'password': {'type': 'string', 'required': True},
                }
            },
            'sessionid': {
                'type': 'string',
                'nullable': True,
                'required': False,
            },
            'base64': {
                'type': 'boolean',
                'nullable': False,
                'required': False,
            },
            'success': {
                'type': 'dict',
                'schema': {
                    'body': {
                        'type': 'list',
                        'required': False
                    },
                    'status': {'type': 'integer', 'required': True},
                },
            },
            'type': {
                'type': 'string',
                'regex': 'post|basic_auth|get|raw_post',
                'required': True
            },
            'url': {
                'type': 'list',
                'required': True,
                'schema': {'type': 'string'}
            },
        }
    },
    'category': {'type': 'string', 'required': True},
    'contributor': {'type': 'string', 'required': True},
    'fingerprint': {
        'type': 'dict',
        'required': True,
        'schema': {
            'body': {'type': 'list', 'required': False},
            'server_header': {'type': 'string', 'required': False},
            'cookie': {
                'type': 'list',
                'required': False,
                'schema': {
                    'type': 'dict'
                },
            },
            'headers': {
                'type': 'list',
                'required': False,
                'schema': {
                    'type': 'dict'
                }
            },
            'status': {'type': 'integer', 'required': True},
            'basic_auth_realm': {
                'type': 'string',
                'nullable': True,
                'required': False,
            },
            'url': {
                'type': 'list',
                'required': True,
                'schema': {'type': 'string'}
            },
        },
    },
    'default_port': {'type': 'integer', 'required': True},
    'name': {'type': 'string', 'required': True},
    'ssl': {'type': 'boolean', 'required': True},
    'references': {'type': 'list', 'required': False},
    'versions': {'type': 'list', 'required': False},
    'protocol': {'type': 'string', 'required': False},
}


def mkcred():
    """
    TODO:
        - move credentials under auth
        - move auth url under auth
        - move success under auth
        - move auth url under auth
        - move type under auth
    """

    parameters = dict()
    auth_types = ['post', 'basic_auth', 'get', 'raw_post']

    def get_data(field, prompt, boolean=False, integer=False):
        result = cli_prompt(prompt).strip()
        if boolean and result.lower() == 'y':
            result = True
        elif boolean:
            result = False

        if integer:
            result = int(result)

        parameters[field] = result

    get_data('contributor', 'Your name or handle: ')
    get_data('name', 'Name of service (JBoss, Tomcat): ')
    get_data('protocol', 'Protocol of service (http, ssh, ftp): ')
    get_data('category', 'Category of service (general, printer, phone): ')
    get_data('default_port', 'Default port: ', integer=True)
    get_data('ssl', 'Does the service use ssl (y/n): ', boolean=True)

    # Fingerprint
    ###############################################################################
    fp = dict()

    # Fingerprint url is confiured as a list so we can have more than one path
    path = cli_prompt('Path to the fingerprint page (/index.php): ')
    path_list = list()
    path_list.append(path)
    fp['url'] = path_list

    fp_status = cli_prompt('HTTP status code of fingerprint (401, 200): ')
    fp_body = cli_prompt('Unique string in the fingerprint page (Welcome to ***): ')
    server_header = cli_prompt('Server header (if unique): ')
    basic_auth_realm = cli_prompt('Basic Auth Realm: ')

    fp['status'] = int(fp_status)
    if fp_body:
        b = list()
        b.append(fp_body)
        fp['body'] = b
    if basic_auth_realm:
        fp['basic_auth_realm'] = basic_auth_realm
    if server_header:
        fp['server_header'] = server_header

    parameters['fingerprint'] = fp

    # Authentication
    ###############################################################################
    auth = dict()
    headers = list()
    auth_urls = list()
    url = cli_prompt('Authentication URL (/login.php): ')
    auth_urls.append(url)
    auth['url'] = auth_urls

    while True:
        t = cli_prompt('Type of authentication method (post, basic_auth, get, raw_post): ')
        if t in auth_types:
            auth['type'] = t
            break
        else:
            print('Invalid auth type')

    if auth['type'] == 'post' or auth['type'] == 'get':
        form = dict()
        form['username'] = cli_prompt('Name of username field: ')
        form['password'] = cli_prompt('Name of password field: ')
        form_params = cli_prompt('Post parameters, query string or raw post (json, xml): ')

        if form_params:
            form_params = unquote_plus(form_params)  # decode the parameters
            for f in form_params.split('&'):
                fname = f.split('=')[0]
                fvalue = f.split('=')[1]
                if fname == form['username'] or fname == form['password']:
                    continue
                else:
                    form[fname] = fvalue

        if auth['type'] == 'raw_post':
            form['raw'] = form_params

        auth[auth['type']] = form
    while True:
        header = cli_prompt('Pleae enter any custom header needed. Hit enter if done or not needed \n Example: Content-Type: application/json: ')
        if len(header) > 0:
            if len(header.split(':')) == 2:
                h = header.split(':')
                header = {h[0]: h[1]}
                headers.append(header)
            else:
                print('Invalid header.  Headers must be in the format "Header_name: header_value"\n')
        else:
            break
    csrf = cli_prompt('Name of csrf field: ')
    if csrf:
        auth['csrf'] = csrf

    sessionid = cli_prompt('Name of session cookie: ')
    if sessionid:
        auth['sessionid'] = sessionid

    creds = list()
    num_creds = cli_prompt('How many default creds for this service (1, 2, 3): ')
    for i in range(0, int(num_creds)):
        user = cli_prompt('Username %i: ' % (i + 1))
        passwd = cli_prompt('Password %i: ' % (i + 1))

        if auth['type'] == 'raw_post':
            raw = cli_prompt('Raw post %i: ' % (i + 1))
            creds.append({'username': user, 'password': passwd, 'raw': raw})
        else:
            creds.append({'username': user, 'password': passwd})

    auth['credentials'] = creds
    auth['headers'] = headers

    success = dict()
    success['status'] = 200
    success['body'] = list()
    success['body'].append(cli_prompt('Unique string in page of a successful login (Logout</a>): '))

    auth['success'] = success
    parameters['auth'] = auth

    print()
    fname = parameters['name'].lower().replace(' ', '_').replace('/', '_') + '.yml'
    print('Writing config to %s' % fname)

    cdir = os.path.join('creds', parameters['protocol'], parameters['category'])
    if not os.path.isdir(cdir):
        os.makedirs(cdir)

    with open(os.path.join(cdir, fname), 'w') as fout:
        fout.write(yaml.dump(parameters, default_flow_style=False))

    print(yaml.dump(parameters, default_flow_style=False))

    changeme.core.validate_cred(parameters, fname, parameters['category'])



import memcache
from .scanner import Scanner


class MemcachedScanner(Scanner):

    def __init__(self, cred, target, username, password, config):
        super(MemcachedScanner, self).__init__(cred, target, config, username, password)

    def _check(self):
        mc = memcache.Client(['%s:%s' % (self.target.host, self.target.port)], debug=0)
        stats = mc.get_stats()
        evidence = "version: %s" % (stats[0][1]['version'])

        return evidence

    def _mkscanner(self, cred, target, u, p, config):
        return MemcachedScanner(cred, target, u, p, config)



__all__ = [
    'database',
    'ftp',
    'http_basic_auth',
    'http_fingerprint',
    'http_get',
    'http_post',
    'http_raw_post',
    'memcached',
    'mongo',
    'mssql',
    'mysql',
    'postgres',
    'redis_scanner',
    'scanner',
    'snmp',
    'ssh_key',
    'ssh',
    'telnet'
]



import responses


class MockResponses:
    tomcat_fp = {
        'method': responses.GET,
        'url': 'http://127.0.0.1:8080/manager/html',
        'status': 401,
        'adding_headers': {
            'Server': 'Apache-Coyote/1.1',
            'WWW-Authenticate': 'Basic realm="Tomcat Manager Application'}
    }

    tomcat_fp_alt = {
        'method': responses.GET,
        'url': 'http://127.0.0.1:8080/tomcat/manager/html',
        'status': 404,
        'adding_headers': {
            'Server': 'Apache-Coyote/1.1',
            'WWW-Authenticate': 'Basic realm="Tomcat Manager Application'}
    }

    tomcat_auth = {
        'method': responses.GET,
        'url': 'http://127.0.0.1:8080/manager/html',
        'status': 200,
        'body': '<font size="+2">Tomcat Web Application Manager</font>',
        'adding_headers': {'Server': 'Apache-Coyote/1.1'}
    }

    jboss_fp = {
        'method': responses.GET,
        'url': 'http://127.0.0.1:8080/admin-console/login.seam',
        'status': 200,
        'body': '<p>Welcome to the JBoss AS 6 Admin Console.</p><input name="javax.faces.ViewState" value="foobar" />',
        'adding_headers': {
            'Server': 'Apache-Coyote/1.1',
            'Set-Cookie': 'JSESSIONID=foobar'
        }
    }

    jboss_fp_no_csrf = {
        'method': responses.GET,
        'url': 'http://127.0.0.1:8080/admin-console/login.seam',
        'status': 200,
        'body': '<p>Welcome to the JBoss AS 6 Admin Console.</p>',
        'adding_headers': {
            'Server': 'Apache-Coyote/1.1',
            'Set-Cookie': 'JSESSIONID=foobar'
        }
    }

    jboss_auth = {
        'method': responses.POST,
        'url': 'http://127.0.0.1:8080/admin-console/login.seam',
        'status': 200,
        'body': '<a>Logout</a>',
        'adding_headers': {'Server': 'Apache-Coyote/1.1'}
    }

    jboss_auth_fail = {
        'method': responses.POST,
        'url': 'http://127.0.0.1:8080/admin-console/login.seam',
        'status': 200,
        'body': 'Fail',
        'adding_headers': {'Server': 'Apache-Coyote/1.1'}
    }

    idrac_fp = {
        'method': responses.GET,
        'url': 'https://127.0.0.1:443/login.html',
        'status': 200,
        'body': '<title>Integrated Dell Remote Access Controller</title>',
        'adding_headers': {
            'Server': 'Mbedthis-Appweb/2.4.2',
            'Content-type': 'text/xml',
            'Set-Cookie': '_appwebSessionId_=dffaac7c4fb4e3c4cbd46d3691aeb40f;',
        },
        'body': '<title>Integrated Dell Remote Access Controller 6 - Express</title>',
    }

    idrac_auth = {
        'method': responses.POST,
        'url': 'https://127.0.0.1:443/data/login',
        'status': 200,
        'body': '<title>Integrated Dell Remote Access Controller</title>',
        'adding_headers': {
            'Server': 'Mbedthis-Appweb/2.4.2',
            'Content-type': 'text/xml',
            'Set-Cookie': '_appwebSessionId_=dffaac7c4fb4e3c4cbd46d3691aeb40f',
        },
        'body': '<? xml version="1.0" encoding="UTF-8"?> <root> <status>ok</status> <authResult>0</authResult> <forwardUrl>index.html</forwardUrl> </root>'
    }

    zabbix_fp = {
        'method': responses.GET,
        'url': 'http://127.0.0.1/zabbix/index.php',
        'status': 200,
        'body': 'by Zabbix SIA',
    }

    zabbix_auth = {
        'method': responses.POST,
        'url': 'http://127.0.0.1/zabbix/index.php',
        'status': 200,
        'body': '<a>Logout</a>',
    }

    zabbix_fail = {
        'method': responses.POST,
        'url': 'http://127.0.0.1/zabbix/index.php',
        'status': 200,
        'body': 'foobar',
    }

    ipcamera_fp = {
        'method': responses.GET,
        'url': 'http://127.0.0.1:81/',
        'status': 200,
        'body': 'GetXml("login.xml?"+param,OnLoginAckOK,OnLoginAckFail);'
    }

    ipcamera_auth = {
        'method': responses.GET,
        'url': 'http://127.0.0.1:81/login.xml',
        'status': 200,
        'body': '<?xml version="1.0" encoding="UTF-8" ?><Result><Success>1</Success><UserLevel>0</UserLevel><UserGroup>Admin</UserGroup></Result>'
    }

    elasticsearch = {
        'method': responses.GET,
        'url': 'http://127.0.0.1:9200/',
        'status': 200,
        'body': """{
  "name" : "foo",
  "cluster_name" : "elasticsearch",
  "cluster_uuid" : "1C4hbDs6TRetjINxrOKBZw",
  "version" : {
    "number" : "5.0.2",
    "build_hash" : "f6b4951",
    "build_date" : "2016-11-24T10:07:18.101Z",
    "build_snapshot" : false,
    "lucene_version" : "6.2.1"
  },
  "tagline" : "You Know, for Search"
}"""
    }




import argparse
from changeme import *
from copy import deepcopy
import mock
from nose.tools import *
from netaddr import IPAddress


cli_args = {'all': False,
            'category': None,
            'contributors': False,
            'debug': True,
            'delay': 500,
            'dump': False,
            'dryrun': False,
            'fingerprint': False,
            'fresh': True,
            'log': None,
            'mkcred': False,
            'name': None,
            'noversion': True,
            'output': None,
            'oa': False,
            'portoverride': False,
            'protocols': 'http',
            'proxy': None,
            'resume': False,
            'shodan_query': None,
            'shodan_key': None,
            'ssl': False,
            'target': '127.0.0.1',
            'threads': 20,
            'timeout': 10,
            'useragent': None,
            'validate': False,
            'verbose': False,}



def test_banner():
    core.banner(version.__version__)

no_args = deepcopy(cli_args)
no_args['target'] = None
@raises(SystemExit)
@mock.patch('argparse.ArgumentParser.parse_args', return_value=argparse.Namespace(**no_args))
def test_no_args(mock_args):
    args = core.parse_args()
    core.init_logging(args['args'].verbose, args['args'].debug, args['args'].log)
    config = core.Config(args['args'], args['parser'])


args = deepcopy(cli_args)
args['target'] = '127.0.0.1'
@mock.patch('argparse.ArgumentParser.parse_args', return_value=argparse.Namespace(**args))
def test_target(mock_args):
    args = core.parse_args()
    core.init_logging(args['args'].verbose, args['args'].debug, args['args'].log)
    config = core.Config(args['args'], args['parser'])


"""
args = deepcopy(cli_args)
args['targets'] = '/etc/hosts'
args['target'] = None
print args
@mock.patch('argparse.ArgumentParser.parse_args', return_value=argparse.Namespace(**args))
def test_targets(mock_args):
    core.Config()
"""

args = deepcopy(cli_args)
args['contributors'] = True
@mock.patch('argparse.ArgumentParser.parse_args', return_value=argparse.Namespace(**args))
def test_contributors(mock_args):
    args = core.parse_args()
    core.init_logging(args['args'].verbose, args['args'].debug, args['args'].log)
    config = core.Config(args['args'], args['parser'])
    creds = core.load_creds(config)
    core.print_contributors(creds)


args = deepcopy(cli_args)
args['dump'] = True
@mock.patch('argparse.ArgumentParser.parse_args', return_value=argparse.Namespace(**args))
def test_print_creds(mock_args):
    args = core.parse_args()
    core.init_logging(args['args'].verbose, args['args'].debug, args['args'].log)
    config = core.Config(args['args'], args['parser'])
    creds = core.load_creds(config)
    core.print_creds(creds)



import logging
import paramiko
from .ssh import SSH
from io import StringIO


class SSHKey(SSH):

    def __init__(self, cred, target, username, key, config):
        super(SSHKey, self).__init__(cred, target, username, key, config)
        self.logger = logging.getLogger('changeme')

    def _check(self):
        fake = StringIO(self.password)
        if "RSA PRIVATE KEY" in self.password:
            key = paramiko.RSAKey.from_private_key(fake)
        elif "DSA PRIVATE KEY" in self.password:
            key = paramiko.DSSKey.from_private_key(fake)

        c = paramiko.SSHClient()
        c.set_missing_host_key_policy(paramiko.MissingHostKeyPolicy())  # ignore unknown hosts
        c.connect(hostname=self.target.host, port=self.target.port, username=self.username, pkey=key)
        stdin, stdout, stderr = c.exec_command('uname -a')
        evidence = stdout.readlines()[0]
        c.close()

        self.password = 'Private Key'
        return evidence

    def _mkscanner(self, cred, target, u, p, config):
        return SSHKey(cred, target, u, p, config)



from pymongo import MongoClient
from .scanner import Scanner
import socket


class Mongodb(Scanner):

    def __init__(self, cred, target, username, password, config):
        super(Mongodb, self).__init__(cred, target, config, username, password)

    def _check(self):
        u_p = ""
        if self.username or self.password:
            u_p = "%s:%s@" % (self.username, self.password)
        client = MongoClient('mongodb://%s%s:%s/' % (u_p, self.target.host, self.target.port))
        dbs = client.database_names()
        server_info = client.server_info()
        evidence = 'Version: %s, databases: %s' % (server_info['version'], ', '.join(dbs))

        return evidence

    def _mkscanner(self, cred, target, u, p, config):
        return Mongodb(cred, target, u, p, config)



from libnmap.parser import NmapParser as np
import logging
from netaddr import IPNetwork
from netaddr.core import AddrFormatError
import re
from os.path import isfile
import shodan
import socket

class Target(object):
    def __init__(self, host=None, port=None, protocol=None, url=None):
        self.host = host
        if port:
            port = re.sub(r'\D','',str(port))
            if 0 < int(port) < 65535:
                self.port = int(port)
            else:
                #just disregard the port for now.
                self.port = None
        else:
            self.port = None
        self.protocol = protocol
        self.url = url
        self.ip = None

    def __eq__(self, other):
        return self.__dict__ == other.__dict__

    def __hash__(self):
        return id(self)

    def __repr__(self):
        return self.__str__()

    def __str__(self):
        target = self

        if self.host:
            target = self.host

        if self.port:
            target += ":%s" % self.port

        if self.protocol:
            target = "%s://" % self.protocol + target

        if self.url:
            target += self.url

        return str(target)

    def get_ip(self):
        if self.ip is None:
            regex = re.compile(r'^\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3}$')
            result = regex.match(self.host)
            if not result:
                self.ip = socket.gethostbyname(self.host)
            else:
                self.ip = self.host

        return self.ip

    @staticmethod
    def parse_target(target):
        logger = logging.getLogger('changeme')
        targets = set()
        if isfile(target):
            try:
                # parse nmap
                report = np.parse_fromfile(target)
                logger.info('Loaded %i hosts from %s' % (len(report.hosts), target))
                for h in report.hosts:
                    for s in h.services:
                        targets.add(Target(host=h.address, port=s.port))
            except:
                # parse text file
                with open(target, 'r') as fin:
                    for line in fin:
                        res = Target._parse_target_string(line)
                        for t in res:
                            targets.add(t)
        else:
            targets = Target._parse_target_string(target)

        return targets

    @staticmethod
    def _parse_target_string(target):
        logger = logging.getLogger('changeme')
        logger.debug('Parsing target %s' % target)
        target = target.strip().rstrip('/')
        targets = set()
        try:
            for ip in IPNetwork(target).iter_hosts(): #(covers IP or cidr) #3,4
                targets.add(Target(host=str(ip)))
        except AddrFormatError:
            if len(target.split(':')) == 3:
                # mysql://127.0.0.1:3306
                protocol = target.split(':')[0]
                host = target.split(':')[1].replace('//', '')
                port = target.split(':')[2]
                targets.add(Target(host=host, port=port, protocol=protocol))
            elif "://" in target:
                # snmp://127.0.0.1
                protocol = target.split(':')[0]
                host = target.split(':')[1].replace('//', '')
                targets.add(Target(host=host, protocol=protocol))
            elif ":" in target:
                # 127.0.0.1:8080
                host = target.split(':')[0]
                port = target.split(':')[1]
                targets.add(Target(host=host, port=port))
            else:
                targets.add(Target(host=target))

        return targets

    @staticmethod
    def get_shodan_targets(config):
        logger = logging.getLogger('changeme')
        targets = set()
        api = shodan.Shodan(config.shodan_key)
        results = api.search(config.shodan_query)
        logger.debug("shodan results: %s" % results)
        for r in results['matches']:
            targets.add(Target(host=r['ip_str']))

        return targets



from changeme.target import Target
import os

"""
1. nmap.xml
2. targets.txt
3. 127.0.0.1
4. 192.168.1.0/24
5. 192.168.59.139:8080
6. snmp://192.168.59.101
7. mysql://192.168.59.101:33306
"""

def test_nmap():
    path = os.path.dirname(os.path.abspath(__file__))
    nmap = os.path.join(path, "tomcat_nmap.xml")
    targets = Target.parse_target(nmap)
    assert len(targets) == 1
    t = targets.pop()
    path = os.path.dirname(os.path.abspath(__file__))
    print("target: %s" % t)
    assert t == Target(host='127.0.0.1', port='8080')


def test_targets_file():
    target = '/tmp/targets.txt'
    with open(target, 'w') as fout:
        fout.write('127.0.0.1\n')
        fout.write('127.0.0.2:8080\n')

    targets = Target.parse_target(target)
    assert len(targets) == 2

    for t in targets:
        if t.host == '127.0.0.1':
            t1(t)
        else:
            t2(t)

    os.remove(target)


def t1(t):
    assert t == Target(host='127.0.0.1')


def t2(t):
    assert t == Target(host='127.0.0.2', port=8080)


def test_ip():
    target = '127.0.0.1'
    targets = Target.parse_target(target)
    assert len(targets) == 1
    t = targets.pop()
    assert t == Target(host=target)
    assert str(t) == target


def test_cidr():
    target = '192.168.1.0/24'
    targets = Target.parse_target(target)
    assert len(targets) == 254

    # TODO explicitly validate the range
    """
    for ip in IPNetwork(target).iter_hosts():
        print str(ip)
        assert Target(host=str(ip)) in targets
    """


def test_ip_port():
    target = '192.168.1.1:8080'
    targets = Target.parse_target(target)
    assert len(targets) == 1
    t = targets.pop()
    assert t == Target(host='192.168.1.1', port='8080')
    assert str(t) == target


def test_proto_ip():
    target = 'snmp://192.168.1.1'
    targets = Target.parse_target(target)
    assert len(targets) == 1

    t = targets.pop()
    assert t == Target(host='192.168.1.1', protocol='snmp')
    assert str(t) == target


def test_proto_ip_port():
    target = 'snmp://192.168.1.1:8080'
    targets = Target.parse_target(target)
    assert len(targets) == 1

    t = targets.pop()
    assert t == Target(host='192.168.1.1', port=8080, protocol='snmp')
    assert str(t) == target


def test_hostname():
    target = 'example.com'
    targets = Target.parse_target(target)
    assert len(targets) == 1

    t = targets.pop()
    assert t == Target(host='example.com')


def test_hostname_proto():
    target = 'http://example.com'
    targets = Target.parse_target(target)
    assert len(targets) == 1

    t = targets.pop()
    assert t == Target(host='example.com', protocol='http')


def test_hostname_proto_port():
    target = 'http://example.com:80'
    targets = Target.parse_target(target)
    assert len(targets) == 1

    t = targets.pop()
    assert t == Target(host='example.com', port='80', protocol='http')



import logging
from netaddr import IPAddress
import socket


class Scanner(object):
    def __init__(self, cred, target, config, username, password):
        self.logger = logging.getLogger('changeme')
        self.cred = cred
        self.target = target
        if self.target.port is None:
            self.target.port = self.cred['default_port']
        self.config = config
        self.username = username
        self.password = password

    def __hash__(self):
        return id(self)

    def scan(self):
        return self.check_success()

    def fingerprint(self):
        if self.target.port is None:
            self.target.port = self.cred['default_port']
        try:
            sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
            sock.settimeout(3)
            result = sock.connect_ex((str(self.target.host), self.target.port))
            sock.shutdown(2)
            if result == 0:
                return True
                self.logger.info('Port %i open' % self.target.port)
            else:
                return False
        except Exception as e:
            self.logger.debug(str(e))
            return False

    def get_scanners(self, creds):
        scanners = list()
        for pair in self.cred['auth']['credentials']:

            scanners.append(self._mkscanner(self.cred, self.target, pair['username'], pair['password'], self.config))
        return scanners


    def check_success(self):
        try:
            evidence = self._check()
            self.logger.critical('[+] Found %s default cred %s:%s at %s' % (self.cred['name'], self.username, self.password, self.target))
            self.logger.debug('%s %s:%s evidence: %s' % (self.target, self.username, self.password, evidence))
            return {'name': self.cred['name'],
                    'username': self.username,
                    'password': self.password,
                    'target': self.target,
                    'evidence': evidence}

        except Exception as e:
            self.logger.info('Invalid %s default cred %s:%s at %s' % (self.cred['name'], self.username, self.password, str(self.target)))
            self.logger.debug('%s Exception: %s' % (type(e).__name__, str(e)))
            return False

    def _check(self):
        raise NotImplementedError("A Scanner class needs to implement a _check method.")

    def __getstate__(self):
        state = self.__dict__
        state['logger'] = None  # Need to clear the logger when serializing otherwise mp.Queue blows up
        return state

    def __setstate__(self, d):
        self.__dict__ = d
        self.logger = logging.getLogger('changeme')

    def __eq__(self, other):
        return self.__dict__ == other.__dict__
        #return (str(self.target) + self.username + self.password) == (other.target + other.username + other.password)



from .database import Database

class MySQL(Database):
    def __init__(self, cred, target, username, password, config):
        super(MySQL, self).__init__(cred, target, username, password, config)
        self.database = ""
        self.query = "select version();"

    def _mkscanner(self, cred, target, u, p, config):
        return MySQL(cred, target, u, p, config)


class Cred(object):
    def __init__(self, cdict):
        self.name = cdict['name']

        # fingerprint

        # auth
        self.credentials = cdict['auth']['credentials']
        self.csrf = cdict['auth']['csrf']
        self.headers = cdict['auth']['headers']

        self.username = cdict['auth']['username']
        self.password = cdict['auth']['password']
        self.b64 = cdict['auth']['base64']
        self.success = cdict['auth']['success']



import argparse
from cerberus import Validator
from changeme.redis_queue import RedisQueue
import logging
from logutils import colorize
import os
import re
import redis
from .report import Report
import requests
from requests import ConnectionError
from requests.packages.urllib3.exceptions import InsecureRequestWarning
from .scan_engine import ScanEngine
from . import schema
import sys
from . import version
import yaml

PERSISTENT_QUEUE = "data.db" # Instantiated in the scan_engine class


def banner(version):
    b = """
 #####################################################
#       _                                             #
#   ___| |__   __ _ _ __   __ _  ___ _ __ ___   ___   #
#  / __| '_ \ / _` | '_ \ / _` |/ _ \ '_ ` _ \ / _ \\  #
# | (__| | | | (_| | | | | (_| |  __/ | | | | |  __/  #
#  \___|_| |_|\__,_|_| |_|\__, |\___|_| |_| |_|\___|  #
#                         |___/                       #
#  v%s                                             #
#  Default Credential Scanner by @ztgrace             #
 #####################################################
    """ % version

    return b


def main():
    print(banner(version.__version__))

    args = parse_args()
    init_logging(args['args'].verbose, args['args'].debug, args['args'].log)
    config = Config(args['args'], args['parser'])
    if not config.noversion:
        check_version()
    creds = load_creds(config)
    s = None

    if config.mkcred:
        schema.mkcred()
        quit()

    if config.contributors:
        print_contributors(creds)
        quit()

    if config.dump:
        print_creds(creds)
        quit()

    logger = logging.getLogger('changeme')

    if not config.validate:
        check_for_interrupted_scan(config)
        s = ScanEngine(creds, config)
        try:
            s.scan()
        except IOError:
            logger.debug('Caught IOError exception')

        report = Report(s.found_q, config.output)
        report.print_results()

        if config.output and ".json" in config.output or config.output and config.oa:
            report.render_json()
        if config.output and ".csv" in config.output or config.output and config.oa:
            report.render_csv()
        if config.output and ".html" in config.output or config.output and config.oa:
            report.render_html()
        if (config.output and not ('json' in config.output or 'csv' in config.output or 'html' in config.output)) and not config.oa:
            logger.error('Only JSON, CSV and HTML are the only supported output types.')


    return s


def init_logging(verbose=False, debug=False, logfile=None):
    """
    Logging levels:
        - Critical: Default credential found
        - Error: error in the program
        - Warning: Verbose data
        - Info: more verbose
        - Debug: Extra info for debugging purposes
    """
    # Set up our logging object
    logger = logging.getLogger('changeme')

    if debug:
        logger.setLevel(logging.DEBUG)
    elif verbose:
        logger.setLevel(logging.INFO)
    else:
        logger.setLevel(logging.WARNING)

    if logfile:
        # Create file handler which logs even debug messages
        #######################################################################
        fh = logging.FileHandler(logfile)

        # create formatter and add it to the handler
        formatter = logging.Formatter(
            '[%(asctime)s][%(levelname)s] %(message)s')
        fh.setFormatter(formatter)
        logger.addHandler(fh)

    # Set up the StreamHandler so we can write to the console
    ###########################################################################
    # create console handler with a higher log level
    ch = colorize.ColorizingStreamHandler(sys.stdout)

    # set custom colorings:
    ch.level_map[logging.DEBUG] = [None, 2, False]
    ch.level_map[logging.INFO] = [None, 'white', False]
    ch.level_map[logging.WARNING] = [None, 'yellow', False]
    ch.level_map[logging.ERROR] = [None, 'red', False]
    ch.level_map[logging.CRITICAL] = [None, 'green', False]
    if debug:
        formatter = logging.Formatter('[%(asctime)s][%(module)s][%(funcName)s] %(message)s', datefmt='%H:%M:%S')
    else:
        formatter = logging.Formatter('[%(asctime)s] %(message)s', datefmt='%H:%M:%S')
    ch.setFormatter(formatter)
    logger.addHandler(ch)

    # Adjust the loggers for requests and urllib3
    logging.getLogger('requests').setLevel(logging.ERROR)
    logging.getLogger('urllib3').setLevel(logging.ERROR)
    requests.packages.urllib3.disable_warnings(InsecureRequestWarning)

    return logger


class Config(object):
    def __init__(self, args, arg_parser):
        # Convert argparse Namespace to a dict and make the keys + values member variables of the config class
        args = vars(args)
        self.target = None
        for key in args:
            setattr(self, key, args[key])

        self._validate_args(arg_parser)

    def _validate_args(self, ap):
        logger = logging.getLogger('changeme')
        if (not self.validate and not self.contributors and not self.dump and not self.shodan_query
            and not self.mkcred and not self.resume) and not self.target:
            ap.print_help()
            quit()

        if self.proxy and re.match('^https?://[0-9\.]+:[0-9]{1,5}$', self.proxy):
            self.proxy = {'http': self.proxy, 'https': self.proxy}
            logger.info('Setting proxy to %s' % self.proxy)
        elif self.proxy:
            logger.error('Invalid proxy, must be http(s)://x.x.x.x:8080')
            sys.exit()

        if self.delay and self.delay != 0:
            if isinstance(self.delay, int) and 0 <= self.delay <= 1000:
                logger.debug('Delay is set to %d milliseconds' % self.delay)
            else:
                logger.error('Invalid delay type. Delay must be an integer between 0 and 1000.  Delay is: %s' %
                             type(self.delay))

        # Drop logging level to INFO to see the fingerprint messages
        if self.fingerprint:
            logger.setLevel(logging.INFO)

        if self.verbose:
            logger.setLevel(logging.INFO)
        if self.debug or self.validate:
            logger.setLevel(logging.DEBUG)

        self.useragent = {'User-Agent': self.useragent} if self.useragent else {}

        if ',' in self.protocols:
            self.protocols = self.protocols.split(',')

        if self.all:
            self.protocols = 'all'

        logger.debug(self.protocols)

        if self.output and which('phantomjs') is None:
            logger.warning('phantomjs is not in your path, screenshots will not work')


    def _file_exists(self, f):
        if not os.path.isfile(f):
            self.logger.error("File %s not found" % f)
            sys.exit()


def parse_args():
    ap = argparse.ArgumentParser(description='Default credential scanner v%s' % version.__version__)
    ap.add_argument('--all', '-a', action='store_true', help='Scan for all protocols', default=False)
    ap.add_argument('--category', '-c', type=str, help='Category of default creds to scan for', default=None)
    ap.add_argument('--contributors', action='store_true', help='Display cred file contributors')
    ap.add_argument('--debug', '-d', action='store_true', help='Debug output')
    ap.add_argument('--delay', '-dl', type=int, help="Specify a delay in milliseconds to avoid 429 status codes default=500", default=500)
    ap.add_argument('--dump', action='store_true', help='Print all of the loaded credentials')
    ap.add_argument('--dryrun', action='store_true', help='Print urls to be scan, but don\'t scan them')
    ap.add_argument('--fingerprint', '-f', action='store_true', help='Fingerprint targets, but don\'t check creds', default=False)
    ap.add_argument('--fresh', action='store_true', help='Flush any previous scans and start fresh', default=False)
    ap.add_argument('--log', '-l', type=str, help='Write logs to logfile', default=None)
    ap.add_argument('--mkcred', action='store_true', help='Make cred file', default=False)
    ap.add_argument('--name', '-n', type=str, help='Narrow testing to the supplied credential name', default=None)
    ap.add_argument('--noversion', action='store_true', help='Don\'t perform a version check', default=False)
    ap.add_argument('--proxy', '-p', type=str, help='HTTP(S) Proxy', default=None)
    ap.add_argument('--output', '-o', type=str, help='Name of result file. File extension determines type (csv, html, json).', default=None)
    ap.add_argument('--oa', action='store_true', help='Output results files in csv, html and json formats', default=False)
    ap.add_argument('--protocols', type=str, help="Comma separated list of protocols to test: http,ssh,ssh_key. Defaults to http.", default='http')
    ap.add_argument('--portoverride', action='store_true', help='Scan all protocols on all specified ports', default=False)
    ap.add_argument('--redishost', type=str, help='Redis server', default='localhost')
    ap.add_argument('--redisport', type=str, help='Redis server', default='6379')
    ap.add_argument('--resume', '-r', action='store_true', help='Resume previous scan', default=False)
    ap.add_argument('--shodan_query', '-q', type=str, help='Shodan query', default=None)
    ap.add_argument('--shodan_key', '-k', type=str, help='Shodan API key', default=None)
    ap.add_argument('--ssl', action='store_true', help='Force cred to SSL and fall back to non-SSL if an SSLError occurs', default=False)
    ap.add_argument('--threads', '-t', type=int, help='Number of threads, default=10', default=10)
    ap.add_argument('--timeout', type=int, help='Timeout in seconds for a request, default=10', default=10)
    ap.add_argument('--useragent', '-ua', type=str, help="User agent string to use", default=None)
    ap.add_argument('--validate', action='store_true', help='Validate creds files', default=False)
    ap.add_argument('--verbose', '-v', action='store_true', help='Verbose output', default=False)

    # Hack to get the help to show up right
    cli = ' '.join(sys.argv)
    if "-h" in cli or "--help" in cli:
        ap.add_argument('target', type=str, help='Target to scan. Can be IP, subnet, hostname, nmap xml file, text file or proto://host:port', default=None)

    # initial parse to see if an option not requiring a target was used
    args, unknown = ap.parse_known_args()
    if not args.dump and not args.contributors and not args.mkcred and not args.resume and not args.shodan_query and not args.validate:
        ap.add_argument('target', type=str, help='Target to scan. Can be IP, subnet, hostname, nmap xml file, text file or proto://host:port', default=None)

    args = ap.parse_args()

    return {'args': args, 'parser': ap}


def get_protocol(filename):
    parts = filename.split(os.path.sep)
    cred_index = 0
    for p in parts:
        if p == 'creds':
            break
        cred_index += 1

    return parts[cred_index + 1]



def load_creds(config):
    # protocol is based off of the directory and category is a field in the cred file. That way you can
    # have default creds across protocols for a single device like a printer
    logger = logging.getLogger('changeme')
    creds = list()
    total_creds = 0
    cred_names = list()
    cred_path = os.path.join(os.path.dirname(os.path.abspath(__file__)), '..', 'creds')
    logger.debug('cred_path: %s' % cred_path)
    protocols = [proto for proto in os.walk(cred_path)][0][1]
    for root, dirs, files in os.walk(cred_path):
        for fname in files:
            f = os.path.join(root, fname)
            protocol = get_protocol(f)
            if is_yaml(f):
                parsed = parse_yaml(f)
                if parsed:
                    if parsed['name'] in cred_names:
                        pass
                    elif validate_cred(parsed, f, protocol):
                        parsed['protocol'] = protocol  # Add the protocol after the schema validation
                        if in_scope(config.name, config.category, parsed, protocols):
                            total_creds += len(parsed['auth']['credentials'])
                            creds.append(parsed)
                            cred_names.append(parsed['name'])
                            logger.debug('Loaded %s' % parsed['name'])

    print('Loaded %i default credential profiles' % len(creds))
    print('Loaded %i default credentials\n' % total_creds)

    creds
    return creds


def validate_cred(cred, f, protocol):
    valid = True
    if protocol == 'http':
        v = Validator()
        valid = v.validate(cred, schema.http_schema)
        for e in v.errors:
            logging.getLogger('changeme').error("[validate_cred] Validation Error: %s, %s - %s" %
                                                (f, e, v.errors[e]))
    # TODO: implement schema validators for other protocols

    return valid


def parse_yaml(f):
    logger = logging.getLogger('changeme')
    with open(f, 'r') as fin:
        raw = fin.read()
        try:
            parsed = yaml.load(raw)
        except Exception as e:
            logger.error("[parse_yaml] %s is not a valid yaml file" % f)
            return None
    return parsed


def is_yaml(f):
    isyaml = False
    try:
        isyaml = os.path.basename(f).split('.')[1] == 'yml'
    except:
        pass
    return isyaml


def in_scope(name, category, cred, protocols):
    add = True

    if name:
        names = name.split(',')
        found = False
        for n in names:
            if n.lower() in cred['name'].lower():
                found = True

        if found is False:
            add = False

    if category and not cred['category'] == category:
        add = False
    elif cred['protocol'] not in protocols:
        add = False

    return add


def print_contributors(creds):
    contributors = set()
    for cred in creds:
        cred_contributors = cred['contributor'].split(', ')
        for c in cred_contributors:
            contributors.add(c)

    for c in version.contributors:
        contributors.add(c)

    print("Thank you to our contributors!")
    for i in sorted(contributors, key=str.lower):
        print(i)
    print()


def print_creds(creds):
    for cred in creds:
        print("\n%s (%s)" % (cred['name'], cred['category']))
        for i in cred['auth']['credentials']:
            print("  - %s:%s" % (i['username'], i['password']))


def check_for_interrupted_scan(config):
    logger = logging.getLogger('changeme')
    if config.fresh:
        logger.debug("Forcing a fresh scan")
        remove_queues()
    elif config.resume:
        logger.debug("Resuming previous scan")
        return

    fp = RedisQueue('fingerprint')
    scanners = RedisQueue('scanners')
    fp_qsize = 0
    scanners_qsize = 0
    logger.debug('scanners: %s, fp: %s' % (scanners_qsize, fp_qsize))
    try:
        fp_qsize = fp.qsize()
    except redis.exceptions.ConnectionError:
        pass
    try:
        scanners_qsize = scanners.qsize()
    except redis.exceptions.ConnectionError:
        pass

    if fp_qsize > 0 or scanners_qsize > 0:
        if not prompt_for_resume(config):
            remove_queues()

    if fp_qsize == 0 and scanners_qsize == 0:
        # Clear the found queue if there's no fingerprints or scanners in the queues
        try:
            logger.debug('Clearing found_q')
            found_q = RedisQueue('found_q')
            found_q.delete()
        except Exception as e:
            logger.debug('Exception: %s: %s' % (type(e).__name__, e.__str__().replace('\n', '|')))
            pass



def prompt_for_resume(config):
    logger = logging.getLogger('changeme')
    logger.error('A previous scan was interrupted. Type R to resume or F to start a fresh scan')
    answer = ''
    while not (answer == 'R' or answer == 'F'):
        prompt = '(R/F)> '
        answer = ''
        try:
            answer = raw_input(prompt)
        except NameError:
            answer = input(prompt)

        if answer.upper() == 'F':
            logger.debug("Forcing a fresh scan")
        elif answer.upper() == 'R':
            logger.debug("Resuming previous scan")
            config.resume = True

    return config.resume


def remove_queues():
    logger = logging.getLogger('changeme')
    try:
        os.remove(PERSISTENT_QUEUE)
        logger.debug("%s removed" % PERSISTENT_QUEUE)
    except OSError:
        logger.debug("%s didn't exist" % PERSISTENT_QUEUE)
        pass

    # Clear Redis
    queues = ['fingerprint', 'scanners', 'found_q']
    for q in queues:
        logger.debug('Removing %s RedisQueue' % q)
        r = RedisQueue(q)
        try:
            r.delete()
            logger.debug("%s Redis queue removed" % q)
        except:
            logger.debug("%s Redis queue exception" % q)
            pass


def check_version():
    logger = logging.getLogger('changeme')

    try:
        res = requests.get('https://raw.githubusercontent.com/ztgrace/changeme/master/changeme/version.py', timeout=2)
    except ConnectionError:
        logger.debug("Unable to retrieve latest changeme version.")
        return

    latest = res.text.split('\n')[0].split(' = ')[1].replace("'", '')
    if not version.__version__ == latest:
        logger.warning('Your version of changeme is out of date. Local version: %s, Latest: %s' % (str(version.__version__), latest))


# copied from https://stackoverflow.com/questions/377017/test-if-executable-exists-in-python
def which(program):
    import os

    def is_exe(fpath):
        return os.path.isfile(fpath) and os.access(fpath, os.X_OK)

    fpath, fname = os.path.split(program)
    if fpath:
        if is_exe(program):
            return program
    else:
        for path in os.environ["PATH"].split(os.pathsep):
            path = path.strip('"')
            exe_file = os.path.join(path, program)
            if is_exe(exe_file):
                return exe_file

    return None




from .scanner import Scanner
import telnetlib
import re
import time

class Telnet(Scanner):

    def __init__(self, cred, target, username, password, config):
        super(Telnet, self).__init__(cred, target, config, username, password)

    def _check(self):
        try:
            telnet = telnetlib.Telnet(str(self.target.host))
            timeout_allowed = int(self.cred['auth']['blockingio_timeout'])
            wait_for_pass_prompt = int(self.cred['auth']['telnet_read_timeout'])

            retval = telnet.open(str(self.target.host), int(self.target.port), timeout=timeout_allowed)
            retval._has_poll = False    # telnetlib hackery :)
            banner = telnet.read_until("login: ")
            telnet.write(self.username + "\n")

            password = str(self.password) if self.password else ''

            result = telnet.read_until("Password: ", timeout=wait_for_pass_prompt)
            result = Telnet._trim_string(result)

            if "Password:" in result:
                telnet.write(password + "\n")

            else:
                self.logger.debug("Check closed at: 1")
                telnet.close()
                raise Exception("Telnet credential not found")

            telnet.write(b"ls\n")

            # evidence = '(slow connection, evidence not collected)'
            # try:
            #     evidence = telnet.read_all()
            # except:
            #     pass

            evidence = "(slow connection, evidence not collected)"
            time.sleep(3)
            evidence = telnet.read_very_eager()
            evidence_fp_check = Telnet._trim_string(evidence)

            self.logger.debug("Evidence string returned (stripped): %s" % str(evidence_fp_check))
            evidence_fp_check_as_bytes = ":".join("{:02x}".format(ord(c)) for c in evidence_fp_check)
            self.logger.debug("Evidence string returned (bytes): %s" % str(evidence_fp_check_as_bytes))

            # Remove simple echos or additional password prompt (wrong password)
            if (not evidence_fp_check) or (evidence_fp_check == "ls") or ("Password:" in evidence) or (evidence == ""):
                self.logger.debug("Check closed at: 2")
                telnet.close()
                raise Exception("Telnet credential not found")

            # Remove additional prompts to login - we have a correct username, but incorrect password
            if evidence_fp_check.endswith("login:") or evidence_fp_check.endswith("login: "):
                self.logger.debug("Check closed at: 3")
                telnet.close()
                raise Exception("Telnet credential not found")

            telnet.write("exit\n")
            telnet.close()

            return evidence

        except Exception as e:
            self.logger.debug("Error: %s" % str(e))
            raise e

    @staticmethod
    def _trim_string(str_to_trim):
        return str(str_to_trim).replace(' ','').replace('\s','').replace('\t','').replace('\r','').replace('\n','')

    def _mkscanner(self, cred, target, u, p, config):
        return Telnet(cred, target, u, p, config)



import paramiko
from .scanner import Scanner
import socket


class SSH(Scanner):

    def __init__(self, cred, target, username, password, config):
        super(SSH, self).__init__(cred, target, config, username, password)

    def _check(self):
        c = paramiko.SSHClient()
        c.set_missing_host_key_policy(paramiko.MissingHostKeyPolicy())  # ignore unknown hosts
        c.connect(hostname=self.target.host, port=self.target.port, username=self.username, password=self.password)
        stdin, stdout, stderr = c.exec_command('uname -a')
        evidence = stdout.readlines()[0]
        c.close()

        return evidence

    def _mkscanner(self, cred, target, u, p, config):
        return SSH(cred, target, u, p, config)



from .scanner import Scanner
import sqlalchemy


class Database(Scanner):

    def __init__(self, cred, target, username, password, config):
        super(Database, self).__init__(cred, target, config, username, password)
        self.database = None
        self.query = None

    def _check(self):
        url = "%s://%s:%s@%s:%s/%s" % (self.target.protocol, self.username, self.password, self.target.host, self.target.port, self.database)
        engine = sqlalchemy.create_engine(url, connect_args={'connect_timeout': self.config.timeout})
        c = engine.connect()
        res = c.execute(self.query)

        results = list()
        [results.append(i) for i in res.fetchall()]

        return str(results[0][0])

    def _mkscanner(self, cred, target, u, p, config):
        raise NotImplementedError("A Database class needs to implement a _mkscanner method.")



from .scanner import Scanner
import ftplib


class FTP(Scanner):
    def __init__(self, cred, target, username, password, config):
        super(FTP, self).__init__(cred, target, config, username, password)

    def _check(self):
        ftp = ftplib.FTP()
        ftp.connect(self.target.host, self.target.port)

        ftp.login(self.username, self.password)
        evidence = ftp.retrlines('LIST')
        ftp.quit()

        return evidence

    def _mkscanner(self, cred, target, u, p, config):
        return FTP(cred, target, u, p, config)



from changeme.scanners.http_get import HTTPGetScanner
from requests.auth import HTTPBasicAuth


class HTTPBasicAuthScanner(HTTPGetScanner):
    pass

    def _make_request(self):
        self.logger.debug("Requesting %s" % self.target)
        self.response = self.request.get(self.target,
                                         auth=HTTPBasicAuth(self.username, self.password),
                                         verify=False,
                                         proxies=self.config.proxy,
                                         timeout=self.config.timeout,
                                         headers=self.headers,
                                         cookies=self.cookies)



import redis
import pickle


# based on http://peter-hoffmann.com/2012/python-simple-queue-redis-queue.html
class RedisQueue(object):
    """Simple Queue with Redis Backend"""
    def __init__(self, name, namespace='queue', **redis_kwargs):
        """The default connection parameters are: host='localhost', port=6379, db=0"""
        self.__db= redis.Redis(**redis_kwargs)
        self.key = '%s:%s' %(namespace, name)

    def qsize(self):
        """Return the approximate size of the queue."""
        return self.__db.llen(self.key)

    def empty(self):
        """Return True if the queue is empty, False otherwise."""
        return self.qsize() == 0

    def put(self, item):
        """Put item into the queue."""
        self.__db.rpush(self.key, pickle.dumps(item))

    def get(self, block=True, timeout=None):
        """Remove and return an item from the queue.

        If optional args block is true and timeout is None (the default), block
        if necessary until an item is available."""
        if block:
            item = self.__db.blpop(self.key, timeout=timeout)
        else:
            item = self.__db.lpop(self.key)

        if item:
            item = item[1]
        return pickle.loads(item)

    def get_nowait(self):
        """Equivalent to get(False)."""
        return self.get(False)

    def ping(self):
        self.__db.ping()

    def delete(self):
        self.__db.delete(self.key)
        self.__db.flushdb()



import argparse
from changeme import core
from .core import cli_args
from copy import deepcopy
import logging
import mock



def reset_handlers():
    logger = logging.getLogger('changeme')
    logger.handlers = []
    core.remove_queues()

snmp_args = deepcopy(cli_args)
snmp_args['protocols'] = 'snmp'
snmp_args['name'] = 'publicprivate'
snmp_args['target'] = 'demo.snmplabs.com'
@mock.patch('argparse.ArgumentParser.parse_args', return_value=argparse.Namespace(**snmp_args))
def test_snmp(mock_args):
    reset_handlers()
    se = core.main()
    assert se.found_q.qsize() == 2


snmp_args = deepcopy(cli_args)
snmp_args['name'] = 'publicprivate'
snmp_args['target'] = 'snmp://demo.snmplabs.com'
@mock.patch('argparse.ArgumentParser.parse_args', return_value=argparse.Namespace(**snmp_args))
def test_snmp_proto(mock_args):
    reset_handlers()
    se = core.main()
    assert se.found_q.qsize() == 2
    



import argparse
from changeme import core
from .core import cli_args
from copy import deepcopy
import logging
import mock
import os

logger = logging.getLogger('changeme')

def reset_handlers():
    logger = logging.getLogger('changeme')
    logger.handlers = []
    core.remove_queues()

memcached_args = deepcopy(cli_args)
memcached_args['protocols'] = 'memcached'
memcached_args['target'] = '127.0.0.1'
@mock.patch('argparse.ArgumentParser.parse_args', return_value=argparse.Namespace(**memcached_args))
def memcached(mock_args):
    reset_handlers()
    se = core.main()
    try:
        assert se.found_q.qsize() == 1
    except Exception as e:
        # Raise an assertion error if we're in Travis CI and fail
        if os.environ.get('TRAVIS', None):
            raise e
        # Warn if we're not Travis CI
        else:
            logger.warning('memcached failed')




from .database import Database

class Postgres(Database):
    def __init__(self, cred, target, username, password, config):
        super(Postgres, self).__init__(cred, target, username, password, config)
        self.target.protocol = "postgresql+psycopg2"
        self.database = ""
        self.query = "select version();"

    def _mkscanner(self, cred, target, u, p, config):
        return Postgres(cred, target, u, p, config)


from changeme.scanners.http_get import HTTPGetScanner


class HTTPPostScanner(HTTPGetScanner):

    def __init__(self, cred, target, username, password, config, cookies, csrf):
        super(HTTPPostScanner, self).__init__(cred, target, username, password, config, cookies)
        self.csrf = csrf

    def __reduce__(self):
        return (self.__class__, (self.cred, self.target, self.username, self.password, self.config, self.cookies, self.csrf))

    def _make_request(self):
        self.logger.debug('_make_request')
        self.logger.debug("target: %s" % self.target)
        data = self.render_creds(self.cred, self.csrf)
        self.response = self.request.post(self.target,
                                          data,
                                          verify=False,
                                          proxies=self.config.proxy,
                                          timeout=self.config.timeout,
                                          headers=self.headers,
                                          cookies=self.cookies)



import argparse
from changeme import core
from .core import cli_args
from copy import deepcopy
import logging
import mock



def reset_handlers():
    logger = logging.getLogger('changeme')
    logger.handlers = []
    core.remove_queues()

redis_args = deepcopy(cli_args)
redis_args['protocols'] = 'redis'
redis_args['target'] = '127.0.0.1'
@mock.patch('argparse.ArgumentParser.parse_args', return_value=argparse.Namespace(**redis_args))
def test_redis(mock_args):
    reset_handlers()
    se = core.main()
    assert se.found_q.qsize() == 1

